/*

deblobber.c 2013-10-13 Claude Heiland-Allen <claude@mathr.co.uk>

gcc -std=c99 -Wall -pedantic -Wextra -O3 -fopenmp -lm -lgmp -lmpfr -o deblobber deblobber.c

./deblobber 1280 720 10000 1024 \
  -1.7665131829407844486257580131244137982146185386575345 \
  0.0417232731046392817956957736430615150194156474841886 \
  1e-30 \
  -1.7665131829407844486257580131244137982146185386575345 \
  0.0417232731046392817956957736430615150194156474841886 \
  > deblobber.pgm

*/

#include <complex.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <mpfr.h>

/* magnitude squared (to save a square root in the inner loop) */
static inline double cabs2(complex double x) {
  return creal(x) * creal(x) + cimag(x) * cimag(x);
}

/* compute relative error when adding/subtracting real numbers*/
/* http://en.wikipedia.org/wiki/Loss_of_significance#Loss_of_significant_bits */
static inline double relerr(double x, double y) {
  if ((x > 0 && y < 0) || (x < 0 && y > 0)) {
    double u = fabs(x);
    double v = fabs(y);
    return 1 - fmin(u, v) / fmax(u, v);
  } else {
    return 1;
  }
}

/* compute relative error when adding/subtracting complex numbers*/
static inline double crelerr(complex double x, complex double y) {
  return sqrt(relerr(creal(x), creal(y)) * relerr(cimag(x), cimag(y)));
}

int main(int argc, char **argv) {
  if (argc < 10) {
    fprintf(stderr, "usage: %s width height maxiters precision centerx centery radius refx refy > out.pgm\n", argv[0]);
    return 1;
  }
#define arg_Z(name,index) int name = atoi(argv[index])
#define arg_R(name,index) mpfr_t name; mpfr_init2(name, precision); mpfr_set_str(name, argv[index], 0, MPFR_RNDN)
  arg_Z(width, 1);
  arg_Z(height, 2);
  arg_Z(maxiters, 3);
  arg_Z(precision, 4);
  arg_R(centerx, 5);
  arg_R(centery, 6);
  arg_R(radius, 7);
  arg_R(refx, 8);
  arg_R(refy, 9);
#undef arg_Z
#undef arg_R
  /* compute reference */
  complex double *z = calloc(maxiters, sizeof(*z));
  {
    mpfr_t x, y, xx, yy, xy;
    mpfr_inits2(precision, x, y, xx, yy, xy, (mpfr_ptr) 0);
    mpfr_set_ui(x, 0, MPFR_RNDN);
    mpfr_set_ui(y, 0, MPFR_RNDN);
    for (int k = 0; k < maxiters; ++k) {
      mpfr_sqr(xx, x, MPFR_RNDN);
      mpfr_sqr(yy, y, MPFR_RNDN);
      mpfr_mul(xy, x, y, MPFR_RNDN);
      mpfr_sub(x, xx, yy, MPFR_RNDN);
      mpfr_mul_2ui(y, xy, 1, MPFR_RNDN);
      mpfr_add(x, x, refx, MPFR_RNDN);
      mpfr_add(y, y, refy, MPFR_RNDN);
      z[k] = mpfr_get_d(x, MPFR_RNDN) + I * mpfr_get_d(y, MPFR_RNDN);
    }
    mpfr_clears(x, y, xx, yy, xy, (mpfr_ptr) 0);
  }
  /* compute errors */
  double *es = calloc(width * height, sizeof(*es));
  #pragma omp parallel for
  for (int j = 0; j < height; ++j) {
    for (int i = 0; i < width; ++i) {
      complex double d0;
      {
        mpfr_t px, py;
        mpfr_inits2(precision, px, py, (mpfr_ptr) 0);
        mpfr_set_d(px, (i + 0.5 - width / 2.0) / height, MPFR_RNDN);
        mpfr_set_d(py, (height / 2.0 - 0.5 - j) / height, MPFR_RNDN);
        mpfr_mul(px, px, radius, MPFR_RNDN);
        mpfr_mul(py, py, radius, MPFR_RNDN);
        mpfr_add(px, px, centerx, MPFR_RNDN);
        mpfr_add(py, py, centery, MPFR_RNDN);
        mpfr_sub(px, px, refx, MPFR_RNDN);
        mpfr_sub(py, py, refy, MPFR_RNDN);
        d0 = mpfr_get_d(px, MPFR_RNDN) + I * mpfr_get_d(py, MPFR_RNDN);
        mpfr_clears(px, py, (mpfr_ptr) 0);
      }
      complex double d = d0;
      double e = 1;
      for (int k = 0; k < maxiters; ++k) {
        if (cabs2(z[k] + d)  > 4) {
          break;
        }
        complex double a = 2 * z[k] * d;
        complex double b = d * d;
        complex double c = a + b;
        e *= sqrt(crelerr(a, b) * crelerr(c, d0));
        d = c + d0;
      }
      es[j * width + i] = e;
    }
  }
  /* compute extrema */
  double mine =  1.0 / 0.0;
  double maxe = -1.0 / 0.0;
  {
    int mink = -1;
    for (int k = 0; k < width * height; ++k) {
      if (es[k] < mine) {
        mine = es[k];
        mink = k;
      }
      if (es[k] > maxe) {
        maxe = es[k];
      }
    }
    int i = mink % width;
    int j = mink / width;
    fprintf(stderr, "worst error %f (%f bits) at ", mine, -log2(mine));
    mpfr_t px, py;
    mpfr_inits2(precision, px, py, (mpfr_ptr) 0);
    mpfr_set_d(px, (i + 0.5 - width / 2.0) / height, MPFR_RNDN);
    mpfr_set_d(py, (height / 2.0 - 0.5 - j) / height, MPFR_RNDN);
    mpfr_mul(px, px, radius, MPFR_RNDN);
    mpfr_mul(py, py, radius, MPFR_RNDN);
    mpfr_add(px, px, centerx, MPFR_RNDN);
    mpfr_add(py, py, centery, MPFR_RNDN);
    mpfr_fprintf(stderr, "%Re %Re\n", px, py);
    mpfr_clears(px, py, (mpfr_ptr) 0);
  }
  /* output image */
  printf("P5\n%d %d\n255\n", width, height);
  for (int k = 0; k < width * height; ++k) {
    putchar(fmin(fmax(255 + 0.25 * log2(es[k]), 0), 255));
  }
  /* cleanup */
  free(es);
  free(z);
  mpfr_clears(centerx, centery, radius, refx, refy, (mpfr_ptr) 0);
  return 0;
}
